local FIRST_CATEGORY = 1337
local CURRENT_CATEGORY

-- GLOBALS: FirstLootEver_Items
local unpack, ipairs, tinsert, print, select, time =
	  unpack, ipairs, tinsert, print, select, time
local GetRealmName, GetContainerNumSlots, GetContainerItemID =
	  GetRealmName, GetContainerNumSlots, GetContainerItemID
local BACKPACK_CONTAINER, NUM_BAG_SLOTS =
	  BACKPACK_CONTAINER, NUM_BAG_SLOTS

local gii = GetItemInfo
local iteminfo_cache = {}
local function GetItemInfo(i)
	if iteminfo_cache[i] then return unpack(iteminfo_cache[i]) end
	local retVals = {gii(i)}
	if not retVals[1] then
		-- automatically retry if the client might just not have cached the item yet
		retVals = {gii(i)}
	end
	if not retVals[1] then
		return nil
	end
	iteminfo_cache[i] = retVals
	return unpack(retVals)
end

local addon = CreateFrame("Frame", "FirstLootEver")
local itemGroups = {}
local reverseGroups = {}
function addon:RegisterItem(itemID, itemGroup)
	local g = itemGroups
	for i, k in ipairs(itemGroup) do
		if i > 3 then break end -- maximum of 3 sub-category levels (tab, sub-tab, label)
		g[k] = g[k] or {}
		g = g[k]
	end
	tinsert(g,itemID)
	local tableValue
	for i,k in ipairs(itemGroup) do -- can't use table.concat due to the 3 sub limit - table.concat doesn't allow any limiting of key number
		if i > 3 then break end
		tableValue = (tableValue and tableValue.."." or "")..k
	end
	reverseGroups[itemID] = tableValue
end

local achievementGroups = {}
local reverseAchievements = {}
local reverseBlizzAchievements = {}
function addon:RegisterAchievement(id, name, desc, icon, checkFunc, category)
	if not achievementGroups[category] then
		achievementGroups[category] = {}
		CURRENT_CATEGORY = CURRENT_CATEGORY or (FIRST_CATEGORY+1)
		tinsert(ACHIEVEMENTUI_CATEGORIES,(#ACHIEVEMENTUI_CATEGORIES),{id=CURRENT_CATEGORY, parent=FIRST_CATEGORY, hidden=true})
		reverseBlizzAchievements[CURRENT_CATEGORY] = category
		CURRENT_CATEGORY = CURRENT_CATEGORY + 1
		AchievementFrameCategories_Update()
	end
	tinsert(achievementGroups[category], {id=id, icon=icon, checkFunc=checkFunc, desc=desc, name=name})
	reverseAchievements[id] = {category=category, icon=icon, checkFunc=checkFunc, desc=desc, name=name}
end

function addon:HasAchievement(id)
	if self.savedVar.achievements and self.savedVar.achievements[id] then
		return true
	end
end

function addon:AwardAchievement(id)
	self.savedVar.achievements = self.savedVar.achievements or {}
	self.savedVar.achievements[id] = time()
	self:AnimateOrQueueAchievementPopup(id)
end

local achieveQueue = {}

function addon:AdvanceAchievementQueue()
	if #achieveQueue <= 0 then return end
	local id = achieveQueue[1]
	tremove(achieveQueue,1)
	self:AnimateOrQueueAchievementPopup(id)
end

function addon:AnimateOrQueueAchievementPopup(id)
	local data = reverseAchievements[id]
	local f = AchievementAlertFrame_GetAlertFrame()
	if not f then
		if not AchievementFrame then
			AchievementFrame_LoadUI()
		else
			tinsert(achieveQueue, id)
			print("No frames available :(")
			return
		end
	end
	local fName = f:GetName()
	local display, shieldPts, shieldIcn = _G[fName.."Name"], _G[fName.."ShieldPoints"], _G[fName.."ShieldIcon"]
	if f.guildDisplay then
		f.guildDisplay = nil
		f:SetHeight(88)
		local back = _G[fName.."Background"]
		back:SetTexture("Interface\\AchievemnetFrame\\UI-Achievement-Alert-Background")
		back:SetTexCoord(0,0.605,0,0.703)
		back:SetPoint("TOPLEFT", 0, 0)
		back:SetPoint("BOTTOMRIGHT", 0, 0)
		local iconBrd = _G[fName.."IconOverlay"]
		iconBrd:SetTexture("Interface\\AchievementFrame\UI-Achievement-IconFrame")
		iconBrd:SetTexCoord(0,0.5625,0,0.5625)
		iconBrd:SetPoint("CENTER", -1, 2)
		_G[fName.."Icon"]:SetPoint("TOPLEFT", -26, 16)
		display:SetPoint("BOTTOMLEFT", 72, 36)
		display:SetPoint("BOTTOMRIGHT", -60, 36)
		_G[fName.."Shield"]:SetPoint("TOPRIGHT", -10, -13)
		shieldPts:SetPoint("CENTER", 7, 2)
		shieldPts:SetVertexColor(1,1,1)
		shieldIcn:SetTexCoord(0,0.5,0,0.45)
		local unlock = _[fName.."Unlocked"]
		unlock:SetPoint("TOP", 7, -23)
		unlock:SetText(ACHIEVEMENT_UNLOCKED)
		_G[fName.."GuildName"]:Hide()
		_G[fName.."GuildBorder"]:Hide()
		_G[fName.."GuildBanner"]:Hide()
		f.glow:SetTexture("Interface\\AchievementFrame\\UI-Achievement-Alert-Glow");
		f.glow:SetTexCoord(0, 0.78125, 0, 0.66796875);
		f.shine:SetTexture("Interface\\AchievementFrame\\UI-Achievement-Alert-Glow");
		f.shine:SetTexCoord(0.78125, 0.912109375, 0, 0.28125);
		f.shine:SetPoint("BOTTOMLEFT", 0, 8);
	end
	shieldIcn:SetTexture("Interface\\AchievementFrame\\UI-Achievement-Shields-NoPoints")
	_G[fName.."IconTexture"]:SetTexture(data.icon)
	display:SetText(data.name)
	f.id = nil
	AlertFrame_AnimateIn(f)
	AlertFrame_FixAnchors()
end

function addon:DisplayOrQueueAlert(id)
	print("First looted: "..(select(2,GetItemInfo(id))))
end

function addon:CheckAchievementEligibility(id)
	for avId, data in pairs(reverseAchievements) do
		if data.checkFunc(id) then
			self:AwardAchievement(avId)
		end
	end
end

addon:SetScript("OnEvent", function(self, event, ...)
	if event == "ADDON_LOADED" and (...) == "FirstLootEver" then
		self:UnregisterEvent("ADDON_LOADED")
		FirstLootEver_Items = FirstLootEver_Items or {}
		local r = GetRealmName()
		FirstLootEver_Items[r] = FirstLootEver_Items[r] or {achievements={}}
		self.savedVar = FirstLootEver_Items[r]
	end
	if event == "PLAYER_ENTERING_WORLD" then
		self:UnregisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("BAG_UPDATE")
	end
	if event == "BAG_UPDATE" then
		for bag=BACKPACK_CONTAINER, NUM_BAG_SLOTS do
			for slot=1, GetContainerNumSlots(bag) do
				local id = GetContainerItemID(bag,slot)
				if id then -- not an empty slot
					if not self.savedVar[id] then
						self.savedVar[id] = time()
						addon:DisplayOrQueueAlert(id)
						addon:CheckAchievementEligibility(id)
					end
				end
			end
		end
	end
end)
addon:RegisterEvent("ADDON_LOADED")
addon:RegisterEvent("PLAYER_ENTERING_WORLD")

LoadAddOn("Blizzard_AchievementUI")
hooksecurefunc("AchievementFrameCategories_GetCategoryList", function(t)
	tinsert(t, (#t), {id=FIRST_CATEGORY})
	for blizzID in pairs(reverseBlizzAchievements) do
		tinsert(t, (#t), {id=blizzID, parent=FIRST_CATEGORY, hidden=true})
	end
end)
tinsert(ACHIEVEMENTUI_CATEGORIES,(#ACHIEVEMENTUI_CATEGORIES),{id=FIRST_CATEGORY})

hooksecurefunc("AchievementFrame_LoadUI", function()
	for fID = 1, 2 do
		_G["AchievementAlertFrame"..fID]:HookScript("OnHide", addon.AdvanceAchievementQueue)
	end
end)


-- pre-hook achievement functions to display our category and its contents
local old_GetCategoryInfo = GetCategoryInfo
function GetCategoryInfo(id)
	if id == FIRST_CATEGORY then
		return "FirstLootEver", -1, 0
	elseif reverseBlizzAchievements[id] then
		return reverseBlizzAchievements[id], FIRST_CATEGORY, 0
	else
		return old_GetCategoryInfo(id)
	end
end

local old_GetCategoryNumAchievements = GetCategoryNumAchievements
function GetCategoryNumAchievements(...)
	if (...) == FIRST_CATEGORY then
		local completed, all = 0, 0
		for id in pairs(reverseAchievements) do
			all = all+1
			if addon.savedVar.achievements[id] then
				completed = completed+1
			end
		end
		return all, completed
	elseif reverseBlizzAchievements[(...)] then
		local completed, all = 0, 0
		for _, d in pairs(achievementGroups[reverseBlizzAchievements[(...)]]) do
			all = all+1
			if addon.savedVar.achievements[d.id] then
				completed = completed+1
			end
		end
		return all, completed
	else
		return old_GetCategoryNumAchievements(...)
	end
end

local old_GetAchievementInfo = GetAchievementInfo
function GetAchievementInfo(...)
	if select('#', ...) == 2 then
		local id = select(1,...)
		local achievData
		if id == FIRST_CATEGORY then
			local num = select(2,...)
			for _, data in pairs(achievementGroups) do
				if #data <= num then
					achievData = data[num]
					break
				else
					num = num-#data
				end
			end
		elseif reverseBlizzAchievements[id] then
			achievData = achievementGroups[reverseBlizzAchievements[id]][(select(2,...))]
		else
			return old_GetAchievementInfo(...)
		end
		if not achievData then
			print("Achievement #"..(select(2,...)).." could not be found in category "..id)
			-- prevent the blizz achievement UI from breaking
			return -1, "Invalid achievement", 0, false, nil, nil, nil, "Invalid achievement", 0, "Interface\\Icons\\Spell_ChargeNegative", "", false
		end
		local completionDate = addon.savedVar.achievements[achievData.id]
		if completionDate then
			local dateInfo = date("*t", completionDate)
			return "FirstLootEver_"..achievData.id, achievData.name, 0, true, dateInfo.month, dateInfo.day, (dateInfo.year%100), achievData.desc, 0, achievData.icon, "", false
		else
			return "FirstLootEver_"..achievData.id, achievData.name, 0, false, nil, nil, nil, achievData.desc, 0, achievData.icon, "", false
		end
	else
		local id = (...)
		if type(id) == "string" then
			local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
			if ownID then
				local achievData = reverseAchievements[ownID]
				if not achievData then
					print("Achievement ID "..ownID.." could not be found")
					-- prevent the blizz achievement UI from breaking...again
					return -1, "Invalid achievement", 0, false, nil, nil, nil, "Invalid achievement", 0, "Interface\\Icons\\Spell_ChargeNegative", "", false
				end
				local completionDate = addon.savedVar.achievements[ownID]
				if completionDate then
					local dateInfo = date("*t", completionDate)
					return "FirstLootEver_"..ownID, achievData.name, 0, true, dateInfo.month, dateInfo.day, (dateInfo.year%100), achievData.desc, 0, achievData.icon, "", false
				else
					return "FirstLootEver_"..ownID, achievData.name, 0, false, nil, nil, nil, achievData.desc, 0, achievData.icon, "", false
				end
			end
		end
		return old_GetAchievementInfo(...)
	end
end

local old_IsTrackedAchievement = IsTrackedAchievement
function IsTrackedAchievement(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			return false
		end
	end
	return old_IsTrackedAchievement(id)
end

local old_GetAchievementNumCriteria = GetAchievementNumCriteria
function GetAchievementNumCriteria(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			-- this is temporary. i'll have to think about another way to handle it later.
			-- we just don't want any LUA errors in the blizzard code (for limit must be a number, and they for 1, <this return> it)
			return 0
		end
	end
	return old_GetAchievementNumCriteria(id)
end

local old_GetPreviousAchievement = GetPreviousAchievement
function GetPreviousAchievement(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			-- again temporary to bypass errors
			return nil
		end
	end
	return old_GetPreviousAchievement(id)
end

local old_GetNextAchievement = GetNextAchievement
function GetNextAchievement(id)
	if type(id) == "string" then
		local ownID = tonumber(id:match("FirstLootEver_(%d+)"))
		if ownID then
			--didn't I already tell you? it's temporary. duh.
			return nil
		end
	end
	return old_GetNextAchievement(id)
end